#include<sstream>

#include<boost/regex.hpp>

#include <msgClient.hh>
#include <OPRSclient.hh>
#include <hatpOprsLexer.hh>
#include <ontology-lexer.hh>

//TODO Remove link to libhatp
#include <hatpPlan.hh>
#include <hatpLexer.hh>

using namespace std;
using namespace boost;

static string NAME_OF_HATP	= "";
static string NAME_OF_MSGC	= "";
static string NAME_OF_OPRS	= "";
static string MSGC_SERV 	= "";
static string OPRS_SERV		= "";

static int MSGC_PORT = 5500;
static int OPRS_PORT = 3300;
static int VERBOSE_MODE = 0;

static const regex PlanningRequest(HATPREQ);
static const regex PLANTestOK(PTESTOK);
static const regex PLANTestNO(PTESTNO);
static const regex PLANTestAL(PTESTAL);

int main(int argc, char *argv[]){

	if(argc == 7) {
		NAME_OF_HATP	= argv[1];
		NAME_OF_MSGC	= argv[2];
		MSGC_SERV		= argv[3];
		NAME_OF_OPRS	= argv[4];
		OPRS_SERV		= argv[5];
		VERBOSE_MODE	= atoi(argv[6]);
	} else {
		cerr << "Args error. Usage: NAME_OF_HATP NAME_OF_MSG_CLIENT MSGC_SERV NAME_OF_OPRS OPRS_SERV VERBOSE_MODE" << endl;
		exit(0);
	}
	
	smatch what;
	
	// ----- Connection to msgconnector Server
	msgClient mp;
	mp.connect(NAME_OF_OPRS, MSGC_SERV, MSGC_PORT);
	if(!mp.isConnected()) {
		cerr << "Unable to connect to msgconnector Server... exit." << endl;
		exit(0);
	}

	// ----- Connection To MP-OPRS Server
	OPRSclient mpoprs;
	bool resoprs = false;
	resoprs = mpoprs.connectToHost(NAME_OF_MSGC, OPRS_SERV, OPRS_PORT);
	if(!resoprs) {
		cerr << "Unable to connect to MP-OPRS Server... exit." << endl;
		exit(0);
	}
	long int oprs_id=0;
	
	// ----- Main loop - Message Transfer
	while(mp.isConnected()) {
		usleep(1000); // delay: 1ms
		if(mp.isMessageWaiting()) { // message for msgc ?
			pair<string, string> msg1 = mp.getMessage();
			if(VERBOSE_MODE == 1) cout << "[MSGC]Received from " << msg1.first << ": " << msg1.second << endl;
			string toAnalyse(msg1.second);
			removeFormatting(toAnalyse);
			
			if(regex_match(toAnalyse, what, FindRequest)) {
				ostringstream request, answer;
				request<<"(ONTO_FIND "<<NAME_OF_MSGC<<" "<<oprs_id<<" (ONTO-VAR attributeValue) (. (STATEMENT "<<what[1]<<" "<<what[2]<<" (ONTO-VAR attributeValue)) .) (. .) )";
				answer<<"{\"ONTO-REP\": { \"VALUE\" : [";
				
				mpoprs.sendMsgToMPClient(NAME_OF_OPRS, request.str());

				pair<string, string> reply = mpoprs.getMsgFromMP(true);
				if(reply.second != "NO_NEW_MESSAGE") { // message from oprs ?
					smatch content;
					if(regex_match(reply.second, content, findReply)){
						std::string id(content[1]);
						if(content.size()==4 && content[2]=="ok" && std::atol(id.c_str())==oprs_id){
							string valueField(content[3]);
							boost::sregex_token_iterator valueIterator(valueField.begin(), valueField.end(), findValue), endIterator;
							bool firstRes=true;
							for(; valueIterator!=endIterator; ++valueIterator){
								if(!firstRes) answer<<", ";
								else firstRes=false;
								answer<<"\""<<*valueIterator<<"\"";
							}
						}
					}
				}

				answer << "] }}";
				if(VERBOSE_MODE == 1) cout << "[MSGC]Sent to " << msg1.first << ": " << answer.str() << endl;
				mp.sendMessage(msg1.first, answer.str());
				oprs_id++;
			}else if(regex_match(toAnalyse, what, FindForAgentRequest)) {
				ostringstream request, answer;
				request<<"(ONTO_FIND_FOR_AGENT "<<NAME_OF_MSGC<<" "<<oprs_id<<" "<<what[1]<<" (ONTO-VAR attributeValue) (. (STATEMENT "<<what[2]<<" "<<what[3]<<" (ONTO-VAR attributeValue)) .) (. .) )";
				answer<<"{\"ONTO-REP\": { \"VALUE\" : [";
				
				mpoprs.sendMsgToMPClient(NAME_OF_OPRS, request.str());

				pair<string, string> reply = mpoprs.getMsgFromMP(true);
				if(reply.second != "NO_NEW_MESSAGE") { // message from oprs ?
					smatch content;
					if(regex_match(reply.second, content, findReply)){
						std::string id(content[1]);
						if(content.size()==4 && content[2]=="ok" && std::atol(id.c_str())==oprs_id){
							string valueField(content[3]);
							boost::sregex_token_iterator valueIterator(valueField.begin(), valueField.end(), findValue), endIterator;
							bool firstRes=true;
							for(; valueIterator!=endIterator; ++valueIterator){
								if(!firstRes) answer<<", ";
								else firstRes=false;
								answer<<"\""<<*valueIterator<<"\"";
							}
						}
					}
				}

				answer << "] }}";
				if(VERBOSE_MODE == 1) cout << "[MSGC]Sent to " << msg1.first << ": " << answer.str() << endl;
				mp.sendMessage(msg1.first, answer.str());
				oprs_id++;
				
			}else if(regex_match(toAnalyse, what, FindAllRequest)) {
				ostringstream request, answer;
				request<<"(DbCopy "<<NAME_OF_MSGC<<" "<<oprs_id<<" (. ";

				string objects(what[1]);
				boost::sregex_token_iterator objectsIterator(objects.begin(), objects.end(), ObjectRegex), endIterator;
				
				for(;objectsIterator!=endIterator; ++objectsIterator){
					boost::smatch object;
					
					string currentObject(*objectsIterator);
					if(regex_match(currentObject, object, ObjectRegex)) {
						request<<"(. "<<object[1]<<" ";
						
						string properties(object[2]);
						boost::sregex_token_iterator propertyIterator(properties.begin(), properties.end(), PropertyRegex), endPropIterator;
						for(;propertyIterator!=endPropIterator; ++propertyIterator){
							request<<*propertyIterator<<" ";
						}
						request<<".) ";
					}
				}
				
				request<<".))";
				if(VERBOSE_MODE == 1) cout << "[OPRS] Sent to " << NAME_OF_OPRS << ": " << request.str() << endl;
				mpoprs.sendMsgToMPClient(NAME_OF_OPRS, request.str());

				pair<string, string> reply = mpoprs.getMsgFromMP(true);
				answer<<"{\"ONTO-REP-ALL\": { \"OBJECTS\" : [ ";
				
				if(reply.second != "NO_NEW_MESSAGE") { // message from oprs ?
					smatch content;
					if(regex_match(reply.second, content, copyAllReply)){
						std::string id(content[1]);
						if(content.size()==4 && content[2]=="ok" && std::atol(id.c_str())==oprs_id){
							string objects(content[3]);
							boost::sregex_token_iterator objectIterator(objects.begin(), objects.end(), propertiesReply), endIterator;
							bool firstRes=true;
							for(; objectIterator!=endIterator; ++objectIterator){
								if(!firstRes) answer<<", ";
								else firstRes=false;
								
								answer<<"{ \"VALUE\": [ ";
								
								string tempObject(*objectIterator); //Temporary
								string object(tempObject.substr(2, tempObject.size()-4)); //To remove "(." and ".)"
								boost::sregex_token_iterator valueIterator(object.begin(), object.end(), findValue), endValIterator;
								bool firstVal=true;
								for(; valueIterator!=endValIterator; ++valueIterator){
									if(!firstVal) answer<<", ";
									else firstVal=false;
									
									answer<<"\""<<*valueIterator<<"\"";
								}
								
								answer<<" ] }";
							}
						}
					}
				}

				answer << " ] }}";
				if(VERBOSE_MODE == 1) cout << "[MSGC]Sent to " << msg1.first << ": " << answer.str() << endl;
				mp.sendMessage(msg1.first, answer.str());
				
				oprs_id++;
				
			}else if(regex_match(toAnalyse, what, FindAllForAgentRequest)) {
				ostringstream request, answer;
				request<<"(DbCopyBM "<<NAME_OF_MSGC<<" "<<oprs_id<<" "<<what[1]<<" (. ";

				string objects(what[2]);
				boost::sregex_token_iterator objectsIterator(objects.begin(), objects.end(), ObjectRegex), endIterator;
				
				for(;objectsIterator!=endIterator; ++objectsIterator){
					boost::smatch object;
					
					string currentObject(*objectsIterator);
					if(regex_match(currentObject, object, ObjectRegex)) {
						request<<"(. "<<object[1]<<" ";
						
						string properties(object[2]);
						boost::sregex_token_iterator propertyIterator(properties.begin(), properties.end(), PropertyRegex), endPropIterator;
						for(;propertyIterator!=endPropIterator; ++propertyIterator){
							request<<*propertyIterator<<" ";
						}
						request<<".) ";
					}
				}
				
				request<<".))";
				if(VERBOSE_MODE == 1) cout << "[OPRS] Sent to " << NAME_OF_OPRS << ": " << request.str() << endl;
				mpoprs.sendMsgToMPClient(NAME_OF_OPRS, request.str());

				pair<string, string> reply = mpoprs.getMsgFromMP(true);
				answer<<"{\"ONTO-REP-ALL\": { \"OBJECTS\" : [ ";
				
				if(reply.second != "NO_NEW_MESSAGE") { // message from oprs ?
					smatch content;
					if(regex_match(reply.second, content, copyAllReply)){
						std::string id(content[1]);
						if(content.size()==4 && content[2]=="ok" && std::atol(id.c_str())==oprs_id){
							string objects(content[3]);
							boost::sregex_token_iterator objectIterator(objects.begin(), objects.end(), propertiesReply), endIterator;
							bool firstRes=true;
							for(; objectIterator!=endIterator; ++objectIterator){
								if(!firstRes) answer<<", ";
								else firstRes=false;
								
								answer<<"{ \"VALUE\": [ ";
								
								string tempObject(*objectIterator); //Temporary
								string object(tempObject.substr(2, tempObject.size()-4)); //To remove "(." and ".)"
								boost::sregex_token_iterator valueIterator(object.begin(), object.end(), findValue), endValIterator;
								bool firstVal=true;
								for(; valueIterator!=endValIterator; ++valueIterator){
									if(!firstVal) answer<<", ";
									else firstVal=false;
									
									answer<<"\""<<*valueIterator<<"\"";
								}
								
								answer<<" ] }";
							}
						}
					}
				}

				answer << " ] }}";
				if(VERBOSE_MODE == 1) cout << "[MSGC]Sent to " << msg1.first << ": " << answer.str() << endl;
				mp.sendMessage(msg1.first, answer.str());
				oprs_id++;
				
			}else if(testInputValidity(toAnalyse)) {
				hatpPlan* plan = new hatpPlan(toAnalyse);
				if(plan->isPlanValid()) {
					ostringstream answer;
					answer << "(HATP-REP (REQ-ID " << plan->getID() << ")";
					answer << "(REPORT " << plan->getReport() << ")";
					answer << "(ACTION-LIST (. ";
					hatpStreams* streams = plan->getStreams();
					if(plan->isReportOK()) {
						vector<hatpNode*> nodes = streams->getNodes();
						map<double, unsigned int> orderedList;
						double countInsert = 0, countStep = 0.0001;
						// hack for Matthieu -> reorder actions on time order
						for(unsigned int i = 0 ; i < nodes.size() ; i++) { 
							countInsert += countStep;
							if(orderedList.find((double)(nodes[i]->getStartTime())) == orderedList.end()) {
								orderedList.insert(make_pair((double)(nodes[i]->getStartTime()), i));
							}
							else orderedList.insert(make_pair((double)(nodes[i]->getStartTime())+countInsert, i));
						}
						for(map<double, unsigned int>::iterator it = orderedList.begin(); it != orderedList.end(); it++) {
							unsigned int i = (*it).second;
							answer << "(ACTION ";
							answer << "(ACTION-ID " << nodes[i]->getID() << ")";
							answer << "(ACTION-NAME " << nodes[i]->getName() << ")";
							answer << "(ACTOR " << nodes[i]->getAgents()[0] << ")";
							answer << "(START-TIME " << nodes[i]->getStartTime() << ")";
							answer << "(END-TIME " << nodes[i]->getEndTime() << ")";
							answer << "(PARAMETERS (. ";
							for(unsigned int j = 0 ; j < nodes[i]->getParameters().size() ; j++) {
								answer << nodes[i]->getParameters()[j] << " ";
							}
							answer << " .)))";
						}
					}
					answer << " .))";
					answer << "(LINK-LIST (. ";
					if(plan->isReportOK()) {
						vector<pair<unsigned int, unsigned int> > streamsLinks = streams->getStreamsCausalLinks();
						for(unsigned int i = 0 ; i < streamsLinks.size() ; i++) {
							answer << "(LINK " << streamsLinks[i].first << " " << streamsLinks[i].second << ")";			
						}
					}
					answer << " .)))";
					mpoprs.sendMsgToMPClient(NAME_OF_OPRS, answer.str());
				}else mpoprs.sendMsgToMPClient(NAME_OF_OPRS, msg1.second);
			}else mpoprs.sendMsgToMPClient(NAME_OF_OPRS, msg1.second);
		}
		pair<string, string> msg2 = mpoprs.getMsgFromMP();
		if(msg2.second != "NO_NEW_MESSAGE") { // message from oprs ?
			if(VERBOSE_MODE == 1) cout << "[OPRS]Received from " << msg2.first << ": " << msg2.second << endl;
			string toAnalyse(msg2.second);
			if(regex_match(toAnalyse, what, PlanningRequest)) { 
				ostringstream request;
				request << "{\"HATP-REQ\": { \"REQ-TYPE\": \"" << what[1] << "\", ";
				request << "\"REQ-ID\": " << what[2] << ", ";
				request << "\"TASK-NAME\": \"" << what[3] << "\", ";
				request << "\"PARAMETERS\": [ ";
				string rparams = what[4];
				char *token, *cpar;
				cpar = new char[rparams.size()+1];
				strcpy (cpar, rparams.c_str());
				token = strtok(cpar," ");
				bool first = true;
				while (token != NULL) {
					if(!first ) request << ", ";
					request << "\"" << token << "\"";
					token = strtok(NULL, " ");
					first = false;
				}
				request << " ]}}";
				mp.sendMessage(NAME_OF_HATP, request.str());
			}
			else mp.sendMessage(NAME_OF_HATP, msg2.second);
		}
	}

}
